#include <stdio.h>
#include <strings.h>
#include <math.h>
#define MAX 100
typedef double TYPE;

/* QR algorithm utilizing reflectors for solving
   the least-square problem "a x = b"
 INPUT:
    int n        : number of rows
    int m        : number of columns
    int idata    : data indicator; you may ignore it.
    int level    : print out level; the larger, the more print out
    TYPE a[n][m] : the matrix with n>=m
    TYPE b[n]    : the right-hand side
    int *ierr    : error indicator
 OUTPUT:
    TYPE b[m]    : the solution of the LSP
*/
  
void qr(int n,int m,int idata,int level,
          TYPE a[][m],TYPE b[],TYPE wksp[][m],int*ierr)
{
   int i,j,k,k_end;
   int isi,iga,square,chkmisfit;
   TYPE sigma,gamma,sum,uta,tmp;
   TYPE a_save[MAX][MAX],b_save[MAX],misfit[MAX];

   if(level>=1)printf("QR: idata=%d\n",idata);
   
   isi=0;
   iga=1;

   square=0;
   if(m<n)
      k_end=m;
   else if(m==n)
      {k_end=m-1; square=1;}
   else
      {printf(" Error from QR: n should be at least m: n=$d m=%d\n",n,m);
      *ierr=1; return; }

   chkmisfit=0;
   if(level>=2 && n<=MAX){
      chkmisfit=1;
      for(i=0;i<n;i++){
         b_save[i]=b[i];
         for(j=0;j<m;j++)
            a_save[i][j]=a[i][j];
         }
      }

   /******** BEGIN: QR factorization ************/
   for(k=0;k<k_end;k++){

      sum=0.;
      for(i=k;i<n;i++)
         sum+=a[i][k]*a[i][k];
      sigma=sqrt(sum);

      if(a[k][k]<0)
         sigma=-sigma;
      a[k][k]+=sigma;

      wksp[isi][k]=-sigma;
      wksp[iga][k]=1./(sigma*a[k][k]);

      for(j=k+1;j<m;j++){
         uta=0.;
         for(i=k;i<n;i++)
            uta+=a[i][k]*a[i][j];
         uta*=wksp[iga][k];
         for(i=k;i<n;i++)
           a[i][j]-=uta*a[i][k];
         }
      }

    if(square){
       wksp[isi][m-1]=a[m-1][m-1]; wksp[iga][m-1]=a[m-1][m-1];}
   /******** EDN:  QR factorization ************/

   if(level>=3){
      printf(" The matrix R:\n");
      for(i=0;i<m;i++){
         for(j=0;j<m;j++){
            if(j<i) printf("          ");
            else if(j==i)
              printf(" %f",wksp[isi][j]);
            else
              printf(" %f",a[i][j]);
            }
         printf("\n");
       } 
   }

   /********** BEGIN: Qb **************/
   for(k=0;k<k_end;k++){
      uta=0.;
      for(i=k;i<n;i++) uta+=a[i][k]*b[i];
      uta*=wksp[iga][k];
      for(i=k;i<n;i++) b[i]-=uta*a[i][k];
      }
   /********** EDN:   Qb **************/

   /********** BEGIN: Rx=Qb ***********/
   for(i=m-1;i>=0;i--){
      for(j=i+1;j<m;j++)
         b[i]-=a[i][j]*b[j];
      b[i]/=wksp[isi][i];
      }
   /********** EDN:   Rx=Qb ***********/

    if(level>=1){
       printf(" x=[%g",b[0]);
       for(i=1;i<m;i++)
          printf(",%g",b[i]);
       printf("]\n");
    }

   if(chkmisfit){
      sum=0.;
      for(i=0;i<n;i++){
         misfit[i]=b_save[i];
         for(j=0;j<m;j++)
            misfit[i]-=a_save[i][j]*b[j];
         sum+=misfit[i];
         }
      printf(" misfit=[%.3g",misfit[0]);
      for(i=1;i<n;i++)
           printf(",%.3g",misfit[i]);
      printf("]\n");
      printf(" sum_mf=%.3g\n",sum);
   }

   return;
}

